"""A kinetic model of the methionine cycle.

See here for more about the methionine cycle:
https://doi.org/10.1021/acssynbio.3c00662

"""

import numpy as np
from jax import numpy as jnp

from enzax.kinetic_model import (
    RateEquationKineticModelStructure,
    RateEquationModel,
)
from enzax.rate_equations import (
    AllostericIrreversibleMichaelisMenten,
    Drain,
    IrreversibleMichaelisMenten,
    ReversibleMichaelisMenten,
)


stoichiometry = {
    "the_drain": {"met-L": 1},
    "MAT1": {"met-L": -1, "atp": -1, "pi": 1, "ppi": 1, "amet": 1},
    "MAT3": {"met-L": -1, "atp": -1, "pi": 1, "ppi": 1, "amet": 1},
    "METH-Gen": {"amet": -1, "ahcys": 1},
    "GNMT1": {"amet": -1, "ahcys": 1, "gly": -1, "sarcs": 1},
    "AHC1": {"ahcys": -1, "hcys-L": 1, "adn": 1},
    "MS1": {"hcys-L": -1, "thf": 1, "met-L": 1, "5mthf": -1},
    "BHMT1": {"hcys-L": -1, "glyb": -1, "met-L": 1, "dmgly": 1},
    "CBS1": {"hcys-L": -1, "ser-L": -1, "cyst-L": 1},
    "MTHFR1": {"5mthf": 1, "mlthf": -1, "nadp": 1, "nadph": -1},
    "PROT1": {"met-L": -1},
}
species = [
    "met-L",
    "atp",
    "pi",
    "ppi",
    "amet",
    "ahcys",
    "gly",
    "sarcs",
    "hcys-L",
    "adn",
    "thf",
    "5mthf",
    "mlthf",
    "glyb",
    "dmgly",
    "ser-L",
    "nadp",
    "nadph",
    "cyst-L",
]
balanced_species = [
    "met-L",
    "amet",
    "ahcys",
    "hcys-L",
    "5mthf",
]
reactions = [
    "the_drain",
    "MAT1",
    "MAT3",
    "METH-Gen",
    "GNMT1",
    "AHC1",
    "MS1",
    "BHMT1",
    "CBS1",
    "MTHFR1",
    "PROT1",
]
parameters = dict(
    log_kcat={
        "MAT1": jnp.log(jnp.array(7.89577)),  # MAT1
        "MAT3": jnp.log(jnp.array(19.9215)),  # MAT3
        "METH-Gen": jnp.log(jnp.array(1.15777)),  # METH-Gen
        "GNMT1": jnp.log(jnp.array(10.5307)),  # GNMT1
        "AHC1": jnp.log(jnp.array(234.284)),  # AHC1
        "MS1": jnp.log(jnp.array(1.77471)),  # MS1
        "BHMT1": jnp.log(jnp.array(13.7676)),  # BHMT1
        "CBS1": jnp.log(jnp.array(7.02307)),  # CBS1
        "MTHFR1": jnp.log(jnp.array(3.1654)),  # MTHFR1
        "PROT1": jnp.log(jnp.array(0.264744)),  # PROT1
    },
    log_enzyme={
        "MAT1": jnp.log(jnp.array(0.000961712)),  # MAT1
        "MAT3": jnp.log(jnp.array(0.00098812)),  # MAT3
        "METH-Gen": jnp.log(jnp.array(0.00103396)),  # METH-Gen
        "GNMT1": jnp.log(jnp.array(0.000983692)),  # GNMT1
        "AHC1": jnp.log(jnp.array(0.000977878)),  # AHC1
        "MS1": jnp.log(jnp.array(0.00105094)),  # MS1
        "BHMT1": jnp.log(jnp.array(0.000996603)),  # BHMT1
        "CBS1": jnp.log(jnp.array(0.00134056)),  # CBS1
        "MTHFR1": jnp.log(jnp.array(0.0010054)),  # MTHFR1
        "PROT1": jnp.log(jnp.array(0.000995525)),  # PROT1
    },
    log_drain={"the_drain": jnp.log(jnp.array(0.000850127))},
    dgf=jnp.array(
        [
            160.953,  # met-L
            -2263.31,  # atp
            -1055.95,  # pi
            -1943.8,  # ppi
            636.255,  # amet
            547.319,  # ahcys
            -161.373,  # gly
            -39.4573,  # sarcs
            44.2,  # hcys-L
            375.758,  # adn
            108.366,  # thf
            223.646,  # 5mthf
            198.009,  # mlthf
            173.094,  # glyb
            49.4547,  # dmgly
            -216.712,  # ser-L
            -2014.52,  # nadp
            -1948.58,  # nadph
            -46.4737,  # cyst-L
        ]
    ),
    log_product_km={
        "AHC1": jnp.log(
            jnp.array([1.06e-05, 5.66e-06])
        ),  # hcys-L AHC1, adn AHC1
    },
    log_substrate_km={
        "MAT1": jnp.log(
            jnp.array([0.000106919, 0.00203015])
        ),  # MAT1 met-L, atp
        "MAT3": jnp.log(jnp.array([0.00113258, 0.00236759])),  # MAT3 met-L atp
        "METH-Gen": jnp.log(jnp.array([9.37e-06])),  # METH-Gen amet
        "GNMT1": jnp.log(
            jnp.array([0.000520015, 0.00253545])
        ),  # GNMT1, amet, gly
        "AHC1": jnp.log(jnp.array([2.32e-05])),  # ahcys AHC1
        "MS1": jnp.log(jnp.array([1.71e-06, 6.94e-05])),  # MS1 hcys-L, 5mthf
        "BHMT1": jnp.log(
            jnp.array([1.98e-05, 0.00845898])
        ),  # BHMT1 hcys-L, glyb
        "CBS1": jnp.log(jnp.array([4.24e-05, 2.83e-06])),  #  CBS1 hcys-L, ser-L
        "MTHFR1": jnp.log(
            jnp.array([8.08e-05, 2.09e-05])
        ),  # MTHFR1 mlthf, nadph
        "PROT1": jnp.log(jnp.array([4.39e-05])),  # PROT1 met-L
    },
    temperature=jnp.array(298.15),
    log_ki={
        "MAT1": jnp.array([jnp.log(0.000346704)]),  # MAT1
        "MAT3": jnp.array([]),
        "METH-Gen": jnp.array([jnp.log(5.56e-06)]),  # METH-Gen
        "GNMT1": jnp.array([jnp.log(5.31e-05)]),  # GNMT1
        "AHC1": jnp.array([]),
        "MS1": jnp.array([]),
        "BHMT1": jnp.array([]),
        "CBS1": jnp.array([]),
        "MTHFR1": jnp.array([]),
        "PROT1": jnp.array([]),
    },
    log_conc_unbalanced=jnp.log(
        jnp.array(
            [
                # dataset1
                0.00131546,  # atp
                0.001,  # pi
                0.000500016,  # ppi
                0.00145177,  # gly
                1.00e-07,  # sarcs
                1.01e-06,  # adn
                2.24e-05,  # thf
                3.15e-06,  # mlthf
                0.00106758,  # glyb
                5.00e-05,  # dmgly
                0.0015873,  # ser-L
                1.22e-06,  # nadp
                0.000245139,  # nadph
                2.24e-06,  # cyst-L
            ]
        )
    ),
    log_tc={
        "MAT3": jnp.array(jnp.log(0.107657)),  # MAT3
        "GNMT1": jnp.array(jnp.log(131.207)),  # GNMT
        "CBS1": jnp.array(jnp.log(1.03452)),  # CBS
        "MTHFR1": jnp.array(jnp.log(0.392035)),  # MTHFR
    },
    log_dc_activator={
        "MAT3": jnp.log(
            jnp.array([0.00059999, 0.000316641])
        ),  # met-L MAT3,  # amet MAT3
        "GNMT1": jnp.log(jnp.array([1.98e-05])),  # amet GNMT1
        "CBS1": jnp.array([]),  # CBS1
        "MTHFR1": jnp.log(jnp.array([2.45e-06])),  # ahcys MTHFR1,
    },
    log_dc_inhibitor={
        "MAT3": jnp.array([]),  # MAT3
        "GNMT1": jnp.log(jnp.array([0.000228576])),  # mlthf GNMT1
        "CBS1": jnp.log(jnp.array([9.30e-05])),  # amet CBS1
        "MTHFR1": jnp.log(jnp.array([1.46e-05])),  # amet MTHFR1
    },
)

structure = RateEquationKineticModelStructure(
    stoichiometry=stoichiometry,
    species=species,
    reactions=reactions,
    balanced_species=balanced_species,
    rate_equations=[
        Drain(sign=1.0),  # met-L source
        IrreversibleMichaelisMenten(  # MAT1
            ix_ki_species=np.array([4], dtype=np.int16),
        ),
        AllostericIrreversibleMichaelisMenten(  # MAT3
            subunits=2,
            ix_allosteric_activators=np.array([0, 4], dtype=np.int16),
        ),
        IrreversibleMichaelisMenten(  # METH
            ix_ki_species=np.array([5], dtype=np.int16)
        ),
        AllostericIrreversibleMichaelisMenten(  # GNMT1
            subunits=4,
            ix_allosteric_inhibitors=np.array([12], dtype=np.int16),
            ix_allosteric_activators=np.array([4], dtype=np.int16),
            ix_ki_species=np.array([5], dtype=np.int16),
        ),
        ReversibleMichaelisMenten(  # AHC
            water_stoichiometry=-1.0,
        ),
        IrreversibleMichaelisMenten(),  # MS
        IrreversibleMichaelisMenten(),  # BHMT
        AllostericIrreversibleMichaelisMenten(  # CBS1
            subunits=2,
            ix_allosteric_inhibitors=np.array([4], dtype=np.int16),
        ),
        AllostericIrreversibleMichaelisMenten(  # MTHFR
            subunits=2,
            ix_allosteric_inhibitors=np.array([4], dtype=np.int16),
            ix_allosteric_activators=np.array([5], dtype=np.int16),
        ),
        IrreversibleMichaelisMenten(),  # PROT
    ],
)
steady_state = jnp.array(
    [
        4.233000e-05,  # met-L
        3.099670e-05,  # amet
        2.170170e-07,  # ahcys
        3.521780e-06,  # hcys
        6.534400e-06,  # 5mthf
    ]
)
model = RateEquationModel(parameters, structure)



import libsbml
from sbmlmath import SBMLMathMLParser
import sympy2jax


def load_sbml(file_path):
    reader = libsbml.SBMLReader()
    doc = reader.readSBML(file_path)
    if doc.getModel() is None:
        raise ValueError("Failed to load the SBML model")
    elif doc.getModel().getNumFunctionDefinitions():
        convert_config = (
            libsbml.SBMLFunctionDefinitionConverter().getDefaultProperties()
        )
        doc.convert(convert_config)
    model = doc.getModel()
    return model


def sbml_to_sympy(model):
    reactions_sbml = model.getListOfReactions()
    reactions_sympy = [
        (
            SBMLMathMLParser().parse_str(
                libsbml.writeMathMLToString(
                    libsbml.parseL3Formula(
                        libsbml.formulaToL3String(r.getKineticLaw().getMath())
                    )
                )
            )
        )
        for r in reactions_sbml
    ]
    return reactions_sympy


def sympy_to_enzax(reactions_sympy):
    sym_module = sympy2jax.SymbolicModule(reactions_sympy)
    return sym_module



import jax
import jax.numpy as jnp
from enzax import sbml
from enzax.kinetic_model import KineticModelStructure, KineticModelSbml
from enzax.steady_state import get_kinetic_model_steady_state

jax.config.update("jax_enable_x64", True)

file_path = "tests/data/exampleode_names.xml"
model_sbml = sbml.load_sbml(file_path)
reactions_sympy = sbml.sbml_to_sympy(model_sbml)
sym_module = sbml.sympy_to_enzax(reactions_sympy)

species = [s.getId() for s in model_sbml.getListOfSpecies()]

balanced_species = [
    b.getId() for b in model_sbml.getListOfSpecies() if not b.boundary_condition
]

reactions = [reaction.getId() for reaction in model_sbml.getListOfReactions()]

stoichiometry = {
    reaction.getId(): {
        r.getSpecies(): -r.getStoichiometry(),
        p.getSpecies(): p.getStoichiometry(),
    }
    for reaction in model_sbml.getListOfReactions()
    for r in reaction.getListOfReactants()
    for p in reaction.getListOfProducts()
}

structure = KineticModelStructure(
    stoichiometry=stoichiometry,
    species=species,
    reactions=reactions,
    balanced_species=balanced_species,
)

parameters_local = {
    p.getId(): p.getValue()
    for r in model_sbml.getListOfReactions()
    for p in r.getKineticLaw().getListOfParameters()
}

parameters_global = {
    p.getId(): p.getValue()
    for p in model_sbml.getListOfParameters()
    if p.constant
}

compartments = {c.getId(): c.volume for c in model_sbml.getListOfCompartments()}

unbalanced_species = {
    u.getId(): u.getInitialConcentration()
    for u in model_sbml.getListOfSpecies()
    if u.boundary_condition
}

para = {
    **parameters_local,
    **parameters_global,
    **compartments,
    **unbalanced_species,
}

kinmodel_sbml = KineticModelSbml(
    parameters=para,
    structure=structure,
    sym_module=sym_module,
)

y0 = jnp.array([2.0, 4])
kinmodel_sbml.flux(y0)

kinmodel_sbml.dcdt(t=1, conc=y0)

guess = jnp.full((2), 0.01)
steady_state = get_kinetic_model_steady_state(kinmodel_sbml, guess)
print(steady_state)



import operator
from typing import Union

from enzax.kinetic_model import (
    RateEquationKineticModelStructure,
    RateEquationModel,
    get_conc,
)
from enzax.steady_state import get_kinetic_model_steady_state
import jax
from jax import numpy as jnp
from jax.flatten_util import ravel_pytree
from jax.scipy.stats import multivariate_normal, norm
from jaxtyping import Array, Float, PyTree, Scalar
import equinox as eqx

FloatArray = Float[Array, " _"]
ParamDict = dict[str, Union[FloatArray, "ParamDict"]]


def prior_from_truth(
    true_params: ParamDict,
    sd: float,
    is_multivariate: PyTree[bool] | None = None,
) -> PyTree:
    """Get a PyTree of priors from a PyTree of true values and an sd.

    Params:
    -------
    true_params: PyTree
        A PyTree of true parameter values.
    sd: float
        The standard deviation of the prior.
    is_multivariate: PyTree[bool] | None
        An optional PyTree of booleans indicating whether each parameter is
        multivariate. If None, all parameters are assumed to be univariate.

    Returns:
    --------
    PyTree
        A PyTree of priors, where each leaf is a tuple of the prior location
        and the prior scale (i.e. the argument sd repeated as appropriate).

    """

    def get_scale_for_leaf(true_leaf: jax.Array, is_mv: bool) -> jax.Array:
        sd_arr = jnp.full_like(true_leaf, sd)
        if is_mv:
            sd_arr = jnp.diag(sd_arr)
        return sd_arr

    if is_multivariate is None:
        is_multivariate = jax.tree.map(lambda leaf: False, true_params)
    scale = jax.tree.map(get_scale_for_leaf, true_params, is_multivariate)
    return pack_locs_and_scales(loc=true_params, scale=scale)


def pack_locs_and_scales(loc: PyTree, scale: PyTree) -> PyTree:
    """Get a Pytree whose leaves are tuples of loc and scale.

     Note that loc and scale must have the same tree structure.

     Params:
     -------
     loc: PyTree
         A PyTree representing location parameters.
     scale: PyTree
         A PyTree representing scale parameters.

     Returns:
     --------
     PyTree
         A PyTree whose leaves are tuples of loc and scale.

     Examples:
     ---------
     >>> loc = {"a": jnp.array([1.0, -1.0]), "b": {"c": jnp.array(0.1)}}
     >>> scale = {"a": jnp.array([[1.0, 2.0], [3.0, 4.0]]), "b": {"c": jnp.array(5.0)}}
     >>> get_prior(loc, scale)
     {'a': (Array([ 1., -1.], dtype=float64),
     Array([[1., 2.],
            [3., 4.]], dtype=float64)),
    'b': {'c': (Array(0.1, dtype=float64, weak_type=True),
      Array(5., dtype=float64, weak_type=True))}}

    """  # noqa: E501
    if not jax.tree.structure(loc) == jax.tree.structure(scale):
        msg = "Prior mean and prior sd must have the same tree structure"
        raise ValueError(msg)
    return jax.tree.transpose(
        outer_treedef=jax.tree.structure(("*", "*")),
        inner_treedef=jax.tree.structure(loc),
        pytree_to_transpose=(loc, scale),
    )


def match_prior_leaf(node):
    """Match a Pytree node representing a prior, i.e. a 2-tuple of jax arrays.

    This makes it possible to use jax.tree.map with a function that takes
    a prior leaf as an argument.

    Params:
    -------
    node: Any
        A Pytree node.

    Returns:
    --------
    bool
        True if the node is a 2-tuple of jax arrays
    """
    return (
        (isinstance(node, tuple))
        and (isinstance(node[0], jax.Array))
        and (len(node) == 2)
    )


def enzax_prior_logdensity(parameters: PyTree, prior: PyTree) -> Scalar:
    """Get the prior log density for the parameters given the prior.

    The distribution used depends on the shape of each prior location and scale
    parameter: if they are the same shape, a normal distribution is used; if
    the location is one dimension lower than the scale, a multivariate normal
    distribution is used.

    Params:
    -------
    parameters: PyTree
        A PyTree representing the parameters.
    prior: PyTree
        A PyTree representing the prior. Each leaf is a tuple of the prior
        location and the prior scale (i.e. the argument sd repeated as
        appropriate

    """

    def prior_logdensity_leaf(param_leaf, prior_leaf):
        loc, scale = prior_leaf
        if loc.shape == scale.shape:
            return norm.logpdf(param_leaf, loc, scale).sum()
        elif len(loc.shape) == len(scale.shape) - 1:
            return multivariate_normal.logpdf(param_leaf, loc, scale).sum()
        else:
            msg = "Prior loc and scale have incompatible shape"
            raise ValueError(msg)

    prior_logdensities = jax.tree.map(
        prior_logdensity_leaf,
        parameters,
        prior,
        is_leaf=match_prior_leaf,
    )
    return jax.tree.reduce(operator.add, prior_logdensities)


def enzax_log_likelihood(conc, enzyme, flux) -> Scalar:
    conc_hat, conc_obs, conc_err = conc
    enz_hat, enz_obs, enz_err = enzyme
    flux_hat, flux_obs, flux_err = flux
    llik_conc = norm.logpdf(
        jnp.log(conc_obs), jnp.log(conc_hat), conc_err
    ).sum()
    llik_enz = norm.logpdf(jnp.log(enz_obs), jnp.log(enz_hat), enz_err).sum()
    llik_flux = norm.logpdf(flux_obs, loc=flux_hat, scale=flux_err).sum()
    return llik_conc + llik_enz + llik_flux


def enzax_log_density(
    free_parameters: PyTree,
    structure: RateEquationKineticModelStructure,
    measurements: PyTree,
    prior: PyTree,
    fixed_parameters: PyTree | None = None,
    guess: Float[Array, " _"] | None = None,
) -> Scalar:
    if guess is None:
        guess = jnp.full((len(structure.balanced_species_ix)), 0.01)
    if fixed_parameters is not None:
        parameters = eqx.combine(free_parameters, fixed_parameters)
    else:
        parameters = free_parameters
    model = RateEquationModel(parameters, structure)
    steady = get_kinetic_model_steady_state(model, guess)
    conc_hat = get_conc(steady, parameters["log_conc_unbalanced"], structure)
    flat_log_enzyme, _ = ravel_pytree(parameters["log_enzyme"])
    enz_hat = jnp.exp(jnp.array(flat_log_enzyme))
    flux_hat = model.flux(steady)
    conc_msts, enz_msts, flux_msts = measurements
    log_prior = enzax_prior_logdensity(free_parameters, prior)
    log_likelihood = enzax_log_likelihood(
        (conc_hat, *conc_msts),
        (enz_hat, *enz_msts),
        (flux_hat, *flux_msts),
    )
    return log_prior + log_likelihood



import equinox as eqx
from jax import numpy as jnp
import numpy as np
from numpy.typing import NDArray
from jaxtyping import PyTree, Scalar

from enzax.rate_equation import ConcArray, RateEquation


class DrainInput(eqx.Module):
    abs_v: Scalar


class Drain(RateEquation):
    """A drain reaction."""

    sign: float

    def get_input(
        self,
        parameters: PyTree,
        reaction_id: str,
        reaction_stoichiometry: NDArray[np.float64],
        species_to_dgf_ix: NDArray[np.int16],
    ):
        return DrainInput(abs_v=jnp.exp(parameters["log_drain"][reaction_id]))

    def __call__(self, conc: ConcArray, drain_input: PyTree) -> Scalar:
        """Get the flux of a drain reaction."""
        return self.sign * drain_input.abs_v



from enzax.rate_equations.michaelis_menten import (
    ReversibleMichaelisMenten,
    IrreversibleMichaelisMenten,
)

from enzax.rate_equations.generalised_mwc import (
    AllostericReversibleMichaelisMenten,
    AllostericIrreversibleMichaelisMenten,
)
from enzax.rate_equations.drain import Drain

AVAILABLE_RATE_EQUATIONS = [
    ReversibleMichaelisMenten,
    IrreversibleMichaelisMenten,
    AllostericReversibleMichaelisMenten,
    AllostericIrreversibleMichaelisMenten,
    Drain,
]



import pytest
import importlib.resources
from tests import data
from enzax import sbml

brusselator_file = importlib.resources.files(data) / "brusselator.xml"
exampleode_file = importlib.resources.files(data) / "exampleode.xml"


@pytest.mark.parametrize(
    "file_path",
    [
        brusselator_file,
        exampleode_file,
    ],
)
def test_load_sbml(file_path):
    sbml.load_sbml(file_path)


@pytest.mark.parametrize(
    "model",
    [
        sbml.load_sbml(brusselator_file),
        sbml.load_sbml(exampleode_file),
    ],
)
def test_sbml_to_sympy(model):
    sbml.sbml_to_sympy(model)


def test_sympy_to_enzax(): ...



"""Code for MCMC-based Bayesian inference on kinetic models."""

import functools
from typing import Callable, TypedDict, Unpack

import blackjax
import jax
import jax.numpy as jnp
from jaxtyping import Array, Float, PyTree, ScalarLike


class AdaptationKwargs(TypedDict):
    """Keyword arguments to the blackjax function window_adaptation."""

    initial_step_size: float
    max_num_doublings: int
    is_mass_matrix_diagonal: bool
    target_acceptance_rate: float


@functools.partial(jax.jit, static_argnames=["kernel", "num_samples"])
def _inference_loop(rng_key, kernel, initial_state, num_samples):
    """Run MCMC with blackjax."""

    def one_step(state, rng_key):
        state, info = kernel(rng_key, state)
        return state, (state, info)

    keys = jax.random.split(rng_key, num_samples)
    _, (states, info) = jax.lax.scan(one_step, initial_state, keys)
    return states, info


def run_nuts(
    logdensity_fn: Callable,
    rng_key: Array,
    init_parameters: PyTree,
    num_warmup: int,
    num_samples: int,
    **adapt_kwargs: Unpack[AdaptationKwargs],
):
    """Run the default NUTS algorithm with blackjax."""
    warmup = blackjax.window_adaptation(
        blackjax.nuts,
        logdensity_fn,
        progress_bar=True,
        **adapt_kwargs,
    )
    rng_key, warmup_key = jax.random.split(rng_key)
    (initial_state, tuned_parameters), (_, info, _) = warmup.run(
        warmup_key,
        init_parameters,
        num_steps=num_warmup,  #  type: ignore
    )
    rng_key, sample_key = jax.random.split(rng_key)
    nuts_kernel = blackjax.nuts(logdensity_fn, **tuned_parameters).step
    states, info = _inference_loop(
        sample_key,
        kernel=nuts_kernel,
        initial_state=initial_state,
        num_samples=num_samples,
    )
    return states, info


def ind_prior_from_truth(truth: Float[Array, " _"], sd: ScalarLike):
    """Get a set of independent priors centered at the true parameter values.

    Note that the standard deviation currently has to be the same for
    all parameters.

    """
    return jnp.vstack((truth, jnp.full(truth.shape, sd)))



import equinox as eqx
from jax import numpy as jnp
from jaxtyping import Array, Float, PyTree, Scalar
import numpy as np
from numpy.typing import NDArray

from enzax.rate_equations.michaelis_menten import (
    free_enzyme_ratio_imm,
    free_enzyme_ratio_rmm,
    IrreversibleMichaelisMenten,
    IrreversibleMichaelisMentenInput,
    ReversibleMichaelisMenten,
    ReversibleMichaelisMentenInput,
)


class AllostericIrreversibleMichaelisMentenInput(
    IrreversibleMichaelisMentenInput
):
    dc_inhibitor: Float[Array, " n_inhibitor"]
    dc_activator: Float[Array, " n_activator"]
    tc: Scalar


class AllostericReversibleMichaelisMentenInput(ReversibleMichaelisMentenInput):
    dc_inhibitor: Float[Array, " n_inhibitor"]
    dc_activator: Float[Array, " n_activator"]
    tc: Scalar


def get_allosteric_irreversible_michaelis_menten_input(
    parameters: PyTree,
    reaction_id: str,
    reaction_stoichiometry: NDArray[np.float64],
    species_to_dgf_ix: NDArray[np.int16],
    ci_ix: NDArray[np.int16],
) -> AllostericIrreversibleMichaelisMentenInput:
    ix_substrate = np.argwhere(reaction_stoichiometry < 0.0).flatten()
    return AllostericIrreversibleMichaelisMentenInput(
        kcat=jnp.exp(parameters["log_kcat"][reaction_id]),
        enzyme=jnp.exp(parameters["log_enzyme"][reaction_id]),
        ix_substrate=ix_substrate,
        substrate_kms=jnp.exp(parameters["log_substrate_km"][reaction_id]),
        substrate_stoichiometry=reaction_stoichiometry[ix_substrate],
        ix_ki_species=ci_ix,
        ki=jnp.exp(parameters["log_ki"][reaction_id]),
        dc_inhibitor=jnp.exp(parameters["log_dc_inhibitor"][reaction_id]),
        dc_activator=jnp.exp(parameters["log_dc_activator"][reaction_id]),
        tc=jnp.exp(parameters["log_tc"][reaction_id]),
    )


def get_allosteric_reversible_michaelis_menten_input(
    parameters: PyTree,
    reaction_id: str,
    reaction_stoichiometry: NDArray[np.float64],
    species_to_dgf_ix: NDArray[np.int16],
    ci_ix: NDArray[np.int16],
    water_stoichiometry: float,
) -> AllostericReversibleMichaelisMentenInput:
    ix_reactant = np.argwhere(reaction_stoichiometry != 0.0).flatten()
    ix_substrate = np.argwhere(reaction_stoichiometry < 0.0).flatten()
    ix_product = np.argwhere(reaction_stoichiometry > 0.0).flatten()
    return AllostericReversibleMichaelisMentenInput(
        kcat=jnp.exp(parameters["log_kcat"][reaction_id]),
        enzyme=jnp.exp(parameters["log_enzyme"][reaction_id]),
        substrate_kms=jnp.exp(parameters["log_substrate_km"][reaction_id]),
        product_kms=jnp.exp(parameters["log_product_km"][reaction_id]),
        ki=jnp.exp(parameters["log_ki"][reaction_id]),
        dgf=parameters["dgf"][species_to_dgf_ix][ix_reactant],
        temperature=parameters["temperature"],
        ix_ki_species=ci_ix,
        ix_reactant=ix_reactant,
        ix_substrate=ix_substrate,
        ix_product=ix_product,
        reactant_stoichiometry=reaction_stoichiometry[ix_reactant],
        substrate_stoichiometry=reaction_stoichiometry[ix_substrate],
        product_stoichiometry=reaction_stoichiometry[ix_product],
        water_stoichiometry=water_stoichiometry,
        dc_inhibitor=jnp.exp(parameters["log_dc_inhibitor"][reaction_id]),
        dc_activator=jnp.exp(parameters["log_dc_activator"][reaction_id]),
        tc=jnp.exp(parameters["log_tc"][reaction_id]),
    )


def generalised_mwc_effect(
    conc_inhibitor: Float[Array, " n_inhibition"],
    dc_inhibitor: Float[Array, " n_inhibition"],
    conc_activator: Float[Array, " n_activation"],
    dc_activator: Float[Array, " n_activation"],
    free_enzyme_ratio: Scalar,
    tc: Scalar,
    subunits: int,
) -> Scalar:
    """Get the allosteric effect on a rate.

    The equation is generalised Monod Wyman Changeux model as presented in Popova and Sel'kov 1975: https://doi.org/10.1016/0014-5793(75)80034-2.

    """  # noqa: E501
    qnum = 1 + jnp.sum(conc_inhibitor / dc_inhibitor)
    qdenom = 1 + jnp.sum(conc_activator / dc_activator)
    out = 1.0 / (1 + tc * (free_enzyme_ratio * qnum / qdenom) ** subunits)
    return out


class AllostericIrreversibleMichaelisMenten(IrreversibleMichaelisMenten):
    """A reaction with irreversible Michaelis Menten kinetics and allostery."""

    ix_allosteric_inhibitors: NDArray[np.int16] = eqx.field(
        default_factory=lambda: np.array([], dtype=np.int16)
    )
    ix_allosteric_activators: NDArray[np.int16] = eqx.field(
        default_factory=lambda: np.array([], dtype=np.int16)
    )
    subunits: int = 1

    def get_input(
        self,
        parameters: PyTree,
        reaction_id: str,
        reaction_stoichiometry: NDArray[np.float64],
        species_to_dgf_ix: NDArray[np.int16],
    ):
        return get_allosteric_irreversible_michaelis_menten_input(
            parameters=parameters,
            reaction_id=reaction_id,
            reaction_stoichiometry=reaction_stoichiometry,
            species_to_dgf_ix=species_to_dgf_ix,
            ci_ix=self.ix_ki_species,
        )

    def __call__(
        self,
        conc: Float[Array, " n"],
        aimm_input: AllostericIrreversibleMichaelisMentenInput,
    ) -> Scalar:
        """The flux of an irreversible allosteric Michaelis Menten reaction."""
        fer = free_enzyme_ratio_imm(
            substrate_conc=conc[aimm_input.ix_substrate],
            substrate_km=aimm_input.substrate_kms,
            ki=aimm_input.ki,
            inhibitor_conc=conc[self.ix_ki_species],
            substrate_stoichiometry=aimm_input.substrate_stoichiometry,
        )
        allosteric_effect = generalised_mwc_effect(
            conc_inhibitor=conc[self.ix_allosteric_inhibitors],
            dc_inhibitor=aimm_input.dc_inhibitor,
            dc_activator=aimm_input.dc_activator,
            conc_activator=conc[self.ix_allosteric_activators],
            free_enzyme_ratio=fer,
            tc=aimm_input.tc,
            subunits=self.subunits,
        )
        non_allosteric_rate = super().__call__(conc, aimm_input)
        return non_allosteric_rate * allosteric_effect


class AllostericReversibleMichaelisMenten(ReversibleMichaelisMenten):
    """A reaction with reversible Michaelis Menten kinetics and allostery."""

    ix_allosteric_inhibitors: NDArray[np.int16] = eqx.field(
        default_factory=lambda: np.array([], dtype=np.int16)
    )
    ix_allosteric_activators: NDArray[np.int16] = eqx.field(
        default_factory=lambda: np.array([], dtype=np.int16)
    )
    subunits: int = 1

    def get_input(
        self,
        parameters: PyTree,
        reaction_id: str,
        reaction_stoichiometry: NDArray[np.float64],
        species_to_dgf_ix: NDArray[np.int16],
    ):
        return get_allosteric_reversible_michaelis_menten_input(
            parameters=parameters,
            reaction_id=reaction_id,
            reaction_stoichiometry=reaction_stoichiometry,
            species_to_dgf_ix=species_to_dgf_ix,
            ci_ix=self.ix_ki_species,
            water_stoichiometry=self.water_stoichiometry,
        )

    def __call__(
        self,
        conc: Float[Array, " n"],
        armm_input: AllostericReversibleMichaelisMentenInput,
    ) -> Scalar:
        """The flux of an irreversible allosteric Michaelis Menten reaction."""
        fer = free_enzyme_ratio_rmm(
            substrate_conc=conc[armm_input.ix_substrate],
            product_conc=conc[armm_input.ix_product],
            substrate_kms=armm_input.substrate_kms,
            product_kms=armm_input.product_kms,
            ix_ki_species=conc[self.ix_ki_species],
            ki=armm_input.ki,
            substrate_stoichiometry=armm_input.substrate_stoichiometry,
            product_stoichiometry=armm_input.product_stoichiometry,
        )
        allosteric_effect = generalised_mwc_effect(
            conc_inhibitor=conc[self.ix_allosteric_inhibitors],
            dc_inhibitor=armm_input.dc_inhibitor,
            dc_activator=armm_input.dc_activator,
            conc_activator=conc[self.ix_allosteric_activators],
            free_enzyme_ratio=fer,
            tc=armm_input.tc,
            subunits=self.subunits,
        )
        non_allosteric_rate = super().__call__(conc, armm_input)
        return non_allosteric_rate * allosteric_effect



"""Module for solving steady state problems.

Given a structural kinetic model, a set of parameters and an initial guess, the aim is to find the physiological steady state metabolite concentration and its parameter sensitivities.

"""  # noqa: E501

import diffrax
import equinox as eqx
import lineax as lx
from jaxtyping import Array, Float, PyTree

from enzax.kinetic_model import KineticModel


@eqx.filter_jit()
def get_kinetic_model_steady_state(
    model: KineticModel,
    guess: Float[Array, " n_balanced"],
) -> PyTree:
    """Get the steady state of a kinetic model, using diffrax.

    The better the guess (generally) the faster and more reliable the solving.

    :param model: a KineticModel object

    :param guess: a JAX array of floats. Must have the same length as the
    model's number of balanced species.

    """
    term = diffrax.ODETerm(model.dcdt)
    solver = diffrax.Kvaerno5()
    t0 = 0
    t1 = 900
    dt0 = 0.000001
    max_steps = None
    controller = diffrax.PIDController(
        pcoeff=0.1,
        icoeff=0.3,
        rtol=1e-11,
        atol=1e-11,
    )
    cond_fn = diffrax.steady_state_event()
    event = diffrax.Event(cond_fn)
    adjoint = diffrax.ImplicitAdjoint(
        linear_solver=lx.AutoLinearSolver(well_posed=False)
    )
    sol = diffrax.diffeqsolve(
        term,
        solver,
        t0,
        t1,
        dt0,
        guess,
        max_steps=max_steps,
        stepsize_controller=controller,
        event=event,
        adjoint=adjoint,
    )
    if sol.ys is not None:
        return sol.ys[0]
    else:
        raise ValueError("No steady state found!")



"""Demonstration of how to find a steady state and its gradients with enzax."""

import time
from enzax.kinetic_model import RateEquationModel

import jax
from jax import numpy as jnp

from enzax.examples import methionine
from enzax.steady_state import get_kinetic_model_steady_state
from jaxtyping import PyTree

BAD_GUESS = jnp.full((5,), 0.01)
GOOD_GUESS = jnp.array(
    [
        4.233000e-05,  # met-L
        3.099670e-05,  # amet
        2.170170e-07,  # ahcys
        3.521780e-06,  # hcys
        6.534400e-06,  # 5mthf
    ]
)


def main():
    """Function for testing the steady state solver."""
    model = methionine.model
    # compare good and bad guess
    for guess in [BAD_GUESS, GOOD_GUESS]:

        def get_steady_state_from_params(parameters: PyTree):
            """Get the steady state from just parameters.

            This lets us get the Jacobian wrt (just) the parameters.
            """
            _model = RateEquationModel(parameters, model.structure)
            return get_kinetic_model_steady_state(_model, guess)

        # solve once for jitting
        get_kinetic_model_steady_state(model, GOOD_GUESS)
        jax.jacrev(get_steady_state_from_params)(model.parameters)
        # timer on
        start = time.time()
        conc_steady = get_kinetic_model_steady_state(model, guess)
        jac = jax.jacrev(get_steady_state_from_params)(model.parameters)
        # timer off
        runtime = (time.time() - start) * 1e3
        sv = model.dcdt(jnp.array(0.0), conc=conc_steady)
        flux = model.flux(conc_steady)
        print(f"Results with starting guess {guess}:")
        print(f"\tRun time in milliseconds: {round(runtime, 4)}")
        print(f"\tSteady state concentration: {conc_steady}")
        print(f"\tFlux: {flux}")
        print(f"\tSv: {sv}")
        print(f"\tLog substrate Km Jacobian: {jac['log_substrate_km']}")
        print(f"\tDgf Jacobian: {jac['dgf']}")


if __name__ == "__main__":
    main()



from pathlib import Path
import json
import jax
from jax import numpy as jnp
from jaxtyping import Array, Scalar

from enzax.examples import methionine
from enzax.steady_state import get_kinetic_model_steady_state
from enzax.statistical_modelling import enzax_log_density, prior_from_truth

import functools

jax.config.update("jax_enable_x64", True)
SEED = 1234

HERE = Path(__file__).parent
methionine_pldf_grad_file = HERE / "data" / "expected_methionine_gradient.json"

obs_conc = jnp.array(
    [
        3.99618131e-05,
        1.24186458e-03,
        9.44053469e-04,
        4.72041839e-04,
        2.92625684e-05,
        2.04876101e-07,
        1.37054850e-03,
        9.44053469e-08,
        3.32476221e-06,
        9.53494003e-07,
        2.11467977e-05,
        6.16881926e-06,
        2.97376843e-06,
        1.00785260e-03,
        4.72026734e-05,
        1.49849607e-03,
        1.15174523e-06,
        2.31424323e-04,
        2.11467977e-06,
    ],
    dtype=jnp.float64,
)
obs_flux = jnp.array(
    [
        -0.00425181,
        0.03739644,
        0.01397071,
        -0.04154405,
        -0.05396867,
        0.01236334,
        -0.07089178,
        -0.02136595,
        0.00152784,
        -0.02482788,
        -0.01588131,
    ],
    dtype=jnp.float64,
)
obs_enzyme = jnp.array(
    [
        0.00097884,
        0.00100336,
        0.00105027,
        0.00099059,
        0.00096148,
        0.00107917,
        0.00104588,
        0.00138744,
        0.00107483,
        0.0009662,
    ],
    dtype=jnp.float64,
)


class JAXEncoder(json.JSONEncoder):
    def default(self, obj):  # pyright: ignore[reportIncompatibleMethodOverride]
        if isinstance(obj, jnp.ndarray):
            return {
                "_type": "jax_array",
                "data": obj.tolist(),
                "shape": obj.shape,
                "dtype": str(obj.dtype),
            }
        return super().default(obj)


def serialize_jax_dict(jax_dict):
    return json.dumps(jax_dict, cls=JAXEncoder)


def deserialize_jax_dict(file_path):
    def object_hook(dct):
        if "_type" in dct and dct["_type"] == "jax_array":
            return jnp.array(dct["data"], dtype=dct["dtype"])
        return dct

    with open(file_path, "r") as f:
        return json.load(f, object_hook=object_hook)


def test_lp_grad():
    true_parameters = methionine.parameters
    true_model = methionine.model
    default_state_guess = jnp.full((5,), 0.01)
    true_states = get_kinetic_model_steady_state(
        true_model, default_state_guess
    )
    # get true concentration
    true_conc = jnp.zeros(methionine.structure.S.shape[0])
    true_conc = true_conc.at[methionine.structure.balanced_species_ix].set(
        true_states
    )
    true_conc = true_conc.at[methionine.structure.unbalanced_species_ix].set(
        jnp.exp(true_parameters["log_conc_unbalanced"])  # pyright: ignore[reportArgumentType]
    )
    error_conc = jnp.full_like(obs_conc, 0.03)
    error_flux = jnp.full_like(obs_flux, 0.05)
    error_enzyme = jnp.full_like(obs_enzyme, 0.03)
    measurement_values = obs_conc, obs_enzyme, obs_flux
    measurement_errors = error_conc, error_enzyme, error_flux
    measurements = tuple(zip(measurement_values, measurement_errors))
    prior = prior_from_truth(true_parameters, sd=0.1)  # pyright: ignore[reportArgumentType]

    posterior_log_density = jax.jit(
        functools.partial(
            enzax_log_density,
            structure=true_model.structure,
            fixed_parameters=None,
            measurements=measurements,
            prior=prior,
            guess=default_state_guess,
        )
    )
    gradient = jax.jacrev(posterior_log_density)(methionine.parameters)
    expected_gradient = deserialize_jax_dict(methionine_pldf_grad_file)
    for k, obs in gradient.items():
        exp = expected_gradient[k]
        if isinstance(obs, Scalar):
            assert jnp.isclose(obs, exp)
        elif isinstance(obs, Array):
            assert jnp.isclose(obs, exp).all()
        elif isinstance(obs, dict):
            for kk in obs.keys():
                if isinstance(obs[kk], list):
                    for o, e in zip(obs[kk], exp[kk]):
                        assert jnp.isclose(o, e).all()
                elif isinstance(obs[kk], Scalar):
                    assert jnp.isclose(obs[kk], exp[kk])
                elif len(obs[kk]) > 0:
                    assert jnp.isclose(obs[kk], exp[kk]).all()


if __name__ == "__main__":
    test_lp_grad()



"""A simple linear kinetic model."""

import numpy as np
from jax import numpy as jnp

from enzax.kinetic_model import (
    RateEquationKineticModelStructure,
    RateEquationModel,
)
from enzax.rate_equations import (
    AllostericReversibleMichaelisMenten,
    ReversibleMichaelisMenten,
)


stoichiometry = {
    "r1": {"m1e": -1, "m1c": 1},
    "r2": {"m1c": -1, "m2c": 1},
    "r3": {"m2c": -1, "m2e": 1},
}
reactions = ["r1", "r2", "r3"]
species = ["m1e", "m1c", "m2c", "m2e"]
balanced_species = ["m1c", "m2c"]
rate_equations = [
    AllostericReversibleMichaelisMenten(
        ix_allosteric_activators=np.array([2]), subunits=1
    ),
    AllostericReversibleMichaelisMenten(
        ix_allosteric_inhibitors=np.array([1]), ix_ki_species=np.array([1])
    ),
    ReversibleMichaelisMenten(water_stoichiometry=0.0),
]
structure = RateEquationKineticModelStructure(
    stoichiometry=stoichiometry,
    species=species,
    reactions=reactions,
    balanced_species=balanced_species,
    species_to_dgf_ix=np.array([0, 0, 1, 1]),
    rate_equations=rate_equations,
)
parameters = dict(
    log_substrate_km={
        "r1": jnp.array([0.1]),
        "r2": jnp.array([0.5]),
        "r3": jnp.array([-1.0]),
    },
    log_product_km={
        "r1": jnp.array([-0.2]),
        "r2": jnp.array([0.0]),
        "r3": jnp.array([0.5]),
    },
    log_kcat={
        "r1": jnp.array(-0.1),
        "r2": jnp.array(0.0),
        "r3": jnp.array(0.1),
    },
    dgf=jnp.array([-3.0, -1.0]),
    log_ki={"r1": jnp.array([]), "r2": jnp.array([1.0]), "r3": jnp.array([])},
    temperature=jnp.array(310.0),
    log_enzyme={
        "r1": jnp.log(jnp.array(0.3)),
        "r2": jnp.log(jnp.array(0.2)),
        "r3": jnp.log(jnp.array(0.1)),
    },
    log_conc_unbalanced=jnp.log(jnp.array([0.5, 0.1])),
    log_tc={"r1": jnp.array(-0.2), "r2": jnp.array(0.3)},
    log_dc_activator={"r1": jnp.array([-0.1]), "r2": jnp.array([])},
    log_dc_inhibitor={"r1": jnp.array([]), "r2": jnp.array([0.2])},
)
true_model = RateEquationModel(structure=structure, parameters=parameters)
steady_state = jnp.array([0.43658744, 0.12695706])
model = RateEquationModel(parameters, structure)



"""Demonstration of how to make a Bayesian kinetic model with enzax."""

import functools
import logging
import warnings


import jax
from jax import numpy as jnp
from jax.flatten_util import ravel_pytree

from enzax.examples import methionine
from enzax.kinetic_model import get_conc
from enzax.mcmc import run_nuts
from enzax.steady_state import get_kinetic_model_steady_state
from enzax.statistical_modelling import enzax_log_density, prior_from_truth

import equinox as eqx
from jaxtyping import PyTree

SEED = 1234

jax.config.update("jax_enable_x64", True)


def get_free_params(params: PyTree) -> PyTree:
    """Given a parameter pytree, return the free parameters.

    To get the free and fixed parameter pytrees with the right structure using
    pytree manipulation functions from equinox:

    ```python
    import equinox as eqx
    is_free = eqx.tree_at(
        get_free_params,
        jax.tree.map(lambda _: False, true_parameters),
        replace_fn=lambda _: True,
    )
    free_params, fixed_params = eqx.partition(true_parameters, is_free)
    ```

    """
    return (
        params["log_kcat"]["MAT1"],
        params["temperature"],
        params["dgf"],
    )


def simulate(key, truth, error):
    """Simulate observations from the true model.

    Args:
        key: jax.random key
        truth: tuple of true concentration, log enzyme and flux
        error: tuple of concentration, enzyme and flux error

    """
    key_conc, key_enz, key_flux = jax.random.split(key, num=3)
    true_conc, true_log_enz, true_flux = truth
    conc_err, enz_err, flux_err = error
    return (
        jnp.exp(jnp.log(true_conc) + jax.random.normal(key_conc) * conc_err),
        jnp.exp(true_log_enz + jax.random.normal(key_enz) * enz_err),
        true_flux + jax.random.normal(key_flux) * flux_err,
    )


def main():
    """Demonstrate How to make a Bayesian kinetic model with enzax."""
    true_parameters = methionine.parameters
    true_model = methionine.model
    default_guess = jnp.full((5,), 0.01)
    true_steady = get_kinetic_model_steady_state(true_model, default_guess)
    is_free = eqx.tree_at(
        get_free_params,
        jax.tree.map(lambda _: False, true_parameters),
        replace_fn=lambda _: True,
    )
    free_params, fixed_params = eqx.partition(true_parameters, is_free)
    is_mv = eqx.tree_at(
        lambda params: params["dgf"],
        jax.tree.map(lambda _: False, free_params),
        replace=True,
    )
    prior = prior_from_truth(free_params, sd=0.1, is_multivariate=is_mv)
    # get true concentration, flux and log enzyme
    true_conc = get_conc(
        true_steady,
        true_parameters["log_conc_unbalanced"],
        methionine.structure,
    )
    true_flux = true_model.flux(true_steady)
    true_log_enz_flat, _ = ravel_pytree(true_parameters["log_enzyme"])
    # simulate observations
    conc_err = jnp.full_like(true_conc, 0.03)
    flux_err = jnp.full_like(true_flux, 0.05)
    enz_err = jnp.full_like(true_log_enz_flat, 0.03)
    key = jax.random.key(SEED)
    key_sim, key_nuts = jax.random.split(key, num=2)
    measurement_errors = (conc_err, enz_err, flux_err)
    measurement_values = simulate(
        key=key_sim,
        truth=(true_conc, true_log_enz_flat, true_flux),
        error=measurement_errors,
    )
    measurements = tuple(zip(measurement_values, measurement_errors))
    posterior_log_density = jax.jit(
        functools.partial(
            enzax_log_density,
            structure=true_model.structure,
            fixed_parameters=fixed_params,
            measurements=measurements,
            prior=prior,
            guess=default_guess,
        )
    )
    states, info = run_nuts(
        posterior_log_density,
        key_nuts,
        free_params,
        num_warmup=2,
        num_samples=2,
        initial_step_size=0.0001,
        max_num_doublings=10,
        is_mass_matrix_diagonal=False,
        target_acceptance_rate=0.95,
    )
    if jnp.any(info.is_divergent):
        n_divergent = info.is_divergent.sum()
        msg = f"There were {n_divergent} post-warmup divergent transitions."
        warnings.warn(msg)
    else:
        logging.info("No post-warmup divergent transitions!")
    print("True parameter values vs posterior:")
    for (path, leaf_true), leaf_model in zip(
        jax.tree.leaves_with_path(free_params), jax.tree.leaves(states.position)
    ):
        model_low = jnp.quantile(leaf_model, 0.01, axis=0)
        model_high = jnp.quantile(leaf_model, 0.99, axis=0)
        print(f" {'|'.join(k.key for k in path)}:")
        print(f"  true value: {leaf_true}")
        print(f"  posterior 1%: {model_low}")
        print(f"  posterior 99%: {model_high}")


if __name__ == "__main__":
    main()



"""Unit tests for rate equations."""

import numpy as np
from jax import numpy as jnp
from enzax.rate_equations import (
    AllostericIrreversibleMichaelisMenten,
    AllostericReversibleMichaelisMenten,
    IrreversibleMichaelisMenten,
    ReversibleMichaelisMenten,
)


EXAMPLE_S = np.array([[-1], [1], [0]], dtype=np.float64)
EXAMPLE_CONC = jnp.array([0.5, 0.2, 0.1])
EXAMPLE_PARAMETERS = dict(
    log_substrate_km={"r1": jnp.array([0.1])},
    log_product_km={"r1": jnp.array([-0.2])},
    log_kcat={"r1": jnp.array(-0.1)},
    dgf=jnp.array([-3.0, 1.0]),
    log_ki={"r1": jnp.array([1.0])},
    temperature=jnp.array(310.0),
    log_enzyme={"r1": jnp.log(jnp.array(0.3))},
    log_conc_unbalanced=jnp.array([]),
    log_tc={"r1": jnp.array(-0.2)},
    log_dc_activator={"r1": jnp.array([-0.1])},
    log_dc_inhibitor={"r1": jnp.array([0.2])},
)
EXAMPLE_SPECIES_TO_DGF_IX = np.array([0, 0, 1, 1])


def test_irreversible_michaelis_menten():
    expected_rate = 0.08455524
    f = IrreversibleMichaelisMenten()
    f_input = f.get_input(
        parameters=EXAMPLE_PARAMETERS,
        reaction_id="r1",
        reaction_stoichiometry=EXAMPLE_S[:, 0],
        species_to_dgf_ix=EXAMPLE_SPECIES_TO_DGF_IX,
    )
    rate = f(EXAMPLE_CONC, f_input)
    assert jnp.isclose(rate, expected_rate)


def test_reversible_michaelis_menten():
    expected_rate = 0.04342889
    f = ReversibleMichaelisMenten(
        ix_ki_species=np.array([], dtype=np.int16),
        water_stoichiometry=0.0,
    )
    f_input = f.get_input(
        parameters=EXAMPLE_PARAMETERS,
        reaction_id="r1",
        reaction_stoichiometry=EXAMPLE_S[:, 0],
        species_to_dgf_ix=EXAMPLE_SPECIES_TO_DGF_IX,
    )
    rate = f(EXAMPLE_CONC, f_input)
    assert jnp.isclose(rate, expected_rate)


def test_allosteric_irreversible_michaelis_menten():
    expected_rate = 0.05608589
    f = AllostericIrreversibleMichaelisMenten(
        ix_ki_species=np.array([], dtype=np.int16),
        ix_allosteric_activators=np.array([2], dtype=np.int16),
        subunits=1,
    )
    f_input = f.get_input(
        parameters=EXAMPLE_PARAMETERS,
        reaction_id="r1",
        reaction_stoichiometry=EXAMPLE_S[:, 0],
        species_to_dgf_ix=EXAMPLE_SPECIES_TO_DGF_IX,
    )
    rate = f(EXAMPLE_CONC, f_input)
    assert jnp.isclose(rate, expected_rate)


def test_allosteric_reversible_michaelis_menten():
    expected_rate = 0.03027414
    f = AllostericReversibleMichaelisMenten(
        ix_ki_species=np.array([], dtype=np.int16),
        ix_allosteric_activators=np.array([2], dtype=np.int16),
        subunits=1,
    )
    f_input = f.get_input(
        parameters=EXAMPLE_PARAMETERS,
        reaction_id="r1",
        reaction_stoichiometry=EXAMPLE_S[:, 0],
        species_to_dgf_ix=EXAMPLE_SPECIES_TO_DGF_IX,
    )
    rate = f(EXAMPLE_CONC, f_input)
    assert jnp.isclose(rate, expected_rate)



import equinox as eqx
import numpy as np
from jax import numpy as jnp
from jaxtyping import Array, Float, Int, PyTree, Scalar
from numpy.typing import NDArray

from enzax.rate_equation import RateEquation


class IrreversibleMichaelisMentenInput(eqx.Module):
    kcat: Scalar
    enzyme: Scalar
    ix_ki_species: NDArray[np.int16]
    ki: Float[Array, " n_ki"]
    ix_substrate: NDArray[np.int16]
    substrate_kms: Float[Array, " n_substrate"]
    substrate_stoichiometry: NDArray[np.float64]


def get_irreversible_michaelis_menten_input(
    parameters: PyTree,
    reaction_id: str,
    reaction_stoichiometry: NDArray[np.float64],
    species_to_dgf_ix: NDArray[np.int16],
    ci_ix: NDArray[np.int16],
) -> IrreversibleMichaelisMentenInput:
    ix_substrate = np.argwhere(reaction_stoichiometry < 0.0).flatten()
    return IrreversibleMichaelisMentenInput(
        kcat=jnp.exp(parameters["log_kcat"][reaction_id]),
        enzyme=jnp.exp(parameters["log_enzyme"][reaction_id]),
        ix_substrate=ix_substrate,
        substrate_kms=jnp.exp(parameters["log_substrate_km"][reaction_id]),
        substrate_stoichiometry=reaction_stoichiometry[ix_substrate],
        ix_ki_species=ci_ix,
        ki=jnp.exp(parameters["log_ki"][reaction_id]),
    )


class ReversibleMichaelisMentenInput(eqx.Module):
    kcat: Scalar
    enzyme: Scalar
    ki: Float[Array, " n_ki"]
    substrate_kms: Float[Array, " n_substrate"]
    product_kms: Float[Array, " n_product"]
    dgf: Float[Array, " n_reactant"]
    temperature: Scalar
    ix_ki_species: NDArray[np.int16]
    ix_reactant: NDArray[np.int16]
    ix_substrate: NDArray[np.int16]
    ix_product: NDArray[np.int16]
    reactant_stoichiometry: NDArray[np.float64]
    substrate_stoichiometry: NDArray[np.float64]
    product_stoichiometry: NDArray[np.float64]
    water_stoichiometry: float


def get_reversible_michaelis_menten_input(
    parameters: PyTree,
    reaction_id: str,
    reaction_stoichiometry: NDArray[np.float64],
    species_to_dgf_ix: NDArray[np.int16],
    ci_ix: NDArray[np.int16],
    water_stoichiometry: float,
) -> ReversibleMichaelisMentenInput:
    ix_reactant = np.argwhere(reaction_stoichiometry != 0.0).flatten()
    ix_substrate = np.argwhere(reaction_stoichiometry < 0.0).flatten()
    ix_product = np.argwhere(reaction_stoichiometry > 0.0).flatten()
    return ReversibleMichaelisMentenInput(
        kcat=jnp.exp(parameters["log_kcat"][reaction_id]),
        enzyme=jnp.exp(parameters["log_enzyme"][reaction_id]),
        substrate_kms=jnp.exp(parameters["log_substrate_km"][reaction_id]),
        product_kms=jnp.exp(parameters["log_product_km"][reaction_id]),
        ki=jnp.exp(parameters["log_ki"][reaction_id]),
        dgf=parameters["dgf"][species_to_dgf_ix][ix_reactant],
        temperature=parameters["temperature"],
        ix_ki_species=ci_ix,
        ix_reactant=ix_reactant,
        ix_substrate=ix_substrate,
        ix_product=ix_product,
        reactant_stoichiometry=reaction_stoichiometry[ix_reactant],
        substrate_stoichiometry=reaction_stoichiometry[ix_substrate],
        product_stoichiometry=reaction_stoichiometry[ix_product],
        water_stoichiometry=water_stoichiometry,
    )


def numerator_mm(
    substrate_conc: Float[Array, " n_substrate"],
    substrate_kms: Int[Array, " n_substrate"],
) -> Scalar:
    """Get the product of each substrate's concentration over its km.

    This quantity is the numerator in a Michaelis Menten reaction's rate equation
    """  # Noqa: E501
    return jnp.prod((substrate_conc / substrate_kms))


def get_reversibility(
    reactant_conc: Float[Array, " n_reactant"],
    dgf: Float[Array, " n_reactant"],
    temperature: Scalar,
    reactant_stoichiometry: NDArray[np.float64],
    water_stoichiometry: float,
) -> Scalar:
    """Get the reversibility of a reaction.

    Hard coded water dgf is taken from <http://equilibrator.weizmann.ac.il/metabolite?compoundId=C00001>.

    """
    RT = temperature * 0.008314
    dgf_water = -150.9
    dgr = (
        reactant_stoichiometry.T @ dgf + water_stoichiometry * dgf_water
    ).flatten()
    quotient = (reactant_stoichiometry.T @ jnp.log(reactant_conc)).flatten()
    out = 1.0 - jnp.exp((dgr + RT * quotient) / RT)[0]
    return out


def free_enzyme_ratio_imm(
    substrate_conc: Float[Array, " n_substrate"],
    substrate_km: Float[Array, " n_substrate"],
    ki: Float[Array, " n_ki"],
    inhibitor_conc: Float[Array, " n_ki"],
    substrate_stoichiometry: NDArray[np.float64],
) -> Scalar:
    """Free enzyme ratio for irreversible Michaelis Menten reactions."""
    return 1.0 / (
        jnp.prod(
            ((substrate_conc / substrate_km) + 1)
            ** jnp.abs(substrate_stoichiometry)
        )
        + jnp.sum(inhibitor_conc / ki)
    )


def free_enzyme_ratio_rmm(
    substrate_conc: Float[Array, " n_substrate"],
    product_conc: Float[Array, " n_product"],
    substrate_kms: Float[Array, " n_substrate"],
    product_kms: Float[Array, " n_product"],
    ix_ki_species: Float[Array, " n_ki"],
    ki: Float[Array, " n_ki"],
    substrate_stoichiometry: NDArray[np.float64],
    product_stoichiometry: NDArray[np.float64],
) -> Scalar:
    """The free enzyme ratio for a reversible Michaelis Menten reaction."""
    return 1.0 / (
        -1.0
        + jnp.prod(
            ((substrate_conc / substrate_kms) + 1.0)
            ** jnp.abs(substrate_stoichiometry)
        )
        + jnp.prod(
            ((product_conc / product_kms) + 1.0)
            ** jnp.abs(product_stoichiometry)
        )
        + jnp.sum(ix_ki_species / ki)
    )


class IrreversibleMichaelisMenten(RateEquation):
    """A reaction with irreversible Michaelis Menten kinetics."""

    ix_ki_species: NDArray[np.int16] = eqx.field(
        default_factory=lambda: np.array([], dtype=np.int64)
    )

    def get_input(
        self,
        parameters: PyTree,
        reaction_id: str,
        reaction_stoichiometry: NDArray[np.float64],
        species_to_dgf_ix: NDArray[np.int16],
    ):
        return get_irreversible_michaelis_menten_input(
            parameters=parameters,
            reaction_id=reaction_id,
            reaction_stoichiometry=reaction_stoichiometry,
            species_to_dgf_ix=species_to_dgf_ix,
            ci_ix=self.ix_ki_species,
        )

    def __call__(
        self,
        conc: Float[Array, " n"],
        imm_input: IrreversibleMichaelisMentenInput,
    ) -> Scalar:
        """Get flux of a reaction with irreversible Michaelis Menten kinetics."""  # noqa: E501
        numerator = numerator_mm(
            substrate_conc=conc[imm_input.ix_substrate],
            substrate_kms=imm_input.substrate_kms,
        )
        fer = free_enzyme_ratio_imm(
            substrate_conc=conc[imm_input.ix_substrate],
            substrate_km=imm_input.substrate_kms,
            ki=imm_input.ki,
            inhibitor_conc=conc[imm_input.ix_ki_species],
            substrate_stoichiometry=imm_input.substrate_stoichiometry,
        )
        return imm_input.kcat * imm_input.enzyme * numerator * fer


class ReversibleMichaelisMenten(RateEquation):
    """A reaction with reversible Michaelis Menten kinetics."""

    ix_ki_species: NDArray[np.int16] = eqx.field(
        default_factory=lambda: np.array([], dtype=np.int16)
    )
    water_stoichiometry: float = eqx.field(default_factory=lambda: 0.0)

    def get_input(
        self,
        parameters: PyTree,
        reaction_id: str,
        reaction_stoichiometry: NDArray[np.float64],
        species_to_dgf_ix: NDArray[np.int16],
    ):
        return get_reversible_michaelis_menten_input(
            parameters=parameters,
            reaction_id=reaction_id,
            reaction_stoichiometry=reaction_stoichiometry,
            species_to_dgf_ix=species_to_dgf_ix,
            ci_ix=self.ix_ki_species,
            water_stoichiometry=self.water_stoichiometry,
        )

    def __call__(
        self,
        conc: Float[Array, " n"],
        rmm_input: ReversibleMichaelisMentenInput,
    ) -> Scalar:
        """Get flux of a reaction with reversible Michaelis Menten kinetics.

        :param conc: A 1D array of non-negative numbers representing concentrations of the species that the reaction produces and consumes.

        """  # noqa: E501
        rev = get_reversibility(
            reactant_conc=conc[rmm_input.ix_reactant],
            reactant_stoichiometry=rmm_input.reactant_stoichiometry,
            dgf=rmm_input.dgf,
            temperature=rmm_input.temperature,
            water_stoichiometry=rmm_input.water_stoichiometry,
        )
        numerator = numerator_mm(
            substrate_conc=conc[rmm_input.ix_substrate],
            substrate_kms=rmm_input.substrate_kms,
        )
        fer = free_enzyme_ratio_rmm(
            substrate_conc=conc[rmm_input.ix_substrate],
            product_conc=conc[rmm_input.ix_product],
            ix_ki_species=conc[rmm_input.ix_ki_species],
            substrate_kms=rmm_input.substrate_kms,
            product_kms=rmm_input.product_kms,
            substrate_stoichiometry=rmm_input.substrate_stoichiometry,
            product_stoichiometry=rmm_input.product_stoichiometry,
            ki=rmm_input.ki,
        )
        return rev * rmm_input.kcat * rmm_input.enzyme * numerator * fer



"""Module containing rate equations for enzyme-catalysed reactions."""

from abc import ABC, abstractmethod
from equinox import Module
import numpy as np
from numpy.typing import NDArray

from jaxtyping import Array, Float, PyTree, Scalar


ConcArray = Float[Array, " n"]


class RateEquation(Module, ABC):
    """Abstract definition of a rate equation.

    A rate equation is an equinox [Module](https://docs.kidger.site/equinox/api/module/module/) with a `__call__` method that takes in a 1 dimensional array of concentrations and an arbitrary PyTree of other inputs, returning a scalar value representing a single flux.
    """  # Noqa: E501

    @abstractmethod
    def get_input(
        self,
        parameters: PyTree,
        reaction_id: str,
        reaction_stoichiometry: NDArray[np.float64],
        species_to_dgf_ix: NDArray[np.int16],
    ) -> PyTree: ...

    @abstractmethod
    def __call__(
        self, conc: ConcArray, rate_equation_input: PyTree
    ) -> Scalar: ...



"""Module containing enzax's definition of a kinetic model."""

from abc import ABC, abstractmethod
from typing import Any

import equinox as eqx
import jax.numpy as jnp
import numpy as np
from jax.tree_util import register_pytree_node_class
from jaxtyping import Array, Float, PyTree, ScalarLike, jaxtyped
from numpy.typing import NDArray
from typeguard import typechecked

from enzax.rate_equation import RateEquation


def get_ix_from_list(s: str, list_of_strings: list[str]):
    return next(i for i, si in enumerate(list_of_strings) if si == s)


def get_conc(balanced, log_unbalanced, structure):
    conc = jnp.zeros(structure.S.shape[0])
    conc = conc.at[structure.balanced_species_ix].set(balanced)
    conc = conc.at[structure.unbalanced_species_ix].set(jnp.exp(log_unbalanced))
    return conc


@jaxtyped(typechecker=typechecked)
@register_pytree_node_class
class KineticModelStructure:
    """Structural information about a kinetic model."""

    stoichiometry: dict[str, dict[str, float]]
    species: list[str]
    reactions: list[str]
    balanced_species: list[str]
    unbalanced_species: list[str]
    species_to_dgf_ix: NDArray[np.int16]
    balanced_species_ix: NDArray[np.int16]
    unbalanced_species_ix: NDArray[np.int16]
    S: NDArray[np.float64]

    def __init__(
        self,
        stoichiometry,
        species,
        reactions,
        balanced_species,
        species_to_dgf_ix=None,
    ):
        self.stoichiometry = stoichiometry
        self.species = species
        self.reactions = reactions
        self.balanced_species = balanced_species
        self.unbalanced_species = [
            s for s in species if s not in balanced_species
        ]
        if species_to_dgf_ix is None:
            self.species_to_dgf_ix = np.arange(len(species), dtype=np.int16)
        else:
            self.species_to_dgf_ix = species_to_dgf_ix
        self.balanced_species_ix = np.array(
            [get_ix_from_list(s, species) for s in self.balanced_species],
            dtype=np.int16,
        )
        self.unbalanced_species_ix = np.array(
            [get_ix_from_list(s, species) for s in self.unbalanced_species],
            dtype=np.int16,
        )
        S = np.zeros(shape=(len(species), len(reactions)))
        for ix_reaction, reaction in enumerate(reactions):
            for species_i, coeff in stoichiometry[reaction].items():
                ix_species = get_ix_from_list(species_i, species)
                S[ix_species, ix_reaction] = coeff
        self.S = S.astype(np.float64)

    def tree_flatten(self):
        children = (
            self.stoichiometry,
            self.species,
            self.reactions,
            self.balanced_species,
            self.species_to_dgf_ix,
        )
        aux_data = None
        return children, aux_data

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls(*children)


class RateEquationKineticModelStructure(KineticModelStructure):
    rate_equations: list[RateEquation]

    def __init__(
        self,
        stoichiometry,
        species,
        reactions,
        balanced_species,
        rate_equations,
        species_to_dgf_ix=None,
    ):
        super().__init__(
            stoichiometry,
            species,
            reactions,
            balanced_species,
            species_to_dgf_ix,
        )
        self.rate_equations = rate_equations

    def tree_flatten(self):
        children = (
            self.stoichiometry,
            self.species,
            self.reactions,
            self.balanced_species,
            self.species_to_dgf_ix,
            self.rate_equations,
        )
        aux_data = None
        return children, aux_data


class KineticModel(eqx.Module, ABC):
    """Abstract base class for kinetic models."""

    parameters: PyTree
    structure: KineticModelStructure = eqx.field(static=True)

    @abstractmethod
    def flux(
        self,
        conc_balanced: Float[Array, " n_balanced"],
    ) -> Float[Array, " n"]: ...

    @eqx.filter_jit
    def dcdt(
        self, t: ScalarLike, conc: Float[Array, " n_balanced"], args=None
    ) -> Float[Array, " n_balanced"]:
        """Get the rate of change of balanced species concentrations.

        Note that the signature is as required for a Diffrax vector field function, hence the redundant variable t and the weird name "args".

        :param t: redundant variable representing time.

        :param conc: a one dimensional array of positive floats representing concentrations of balanced species. Must have same size as self.structure.ix_balanced

        """  # Noqa: E501
        v = self.flux(conc)
        sv = self.structure.S @ v
        return jnp.array(sv[self.structure.balanced_species_ix])


class RateEquationModel(KineticModel):
    """A kinetic model that specifies its fluxes using RateEquation objects."""

    structure: RateEquationKineticModelStructure

    def flux(
        self,
        conc_balanced: Float[Array, " n_balanced"],
    ) -> Float[Array, " n"]:
        """Get fluxes from balanced species concentrations.

        :param conc_balanced: a one dimensional array of positive floats representing concentrations of balanced species. Must have same size as self.structure.ix_balanced

        :return: a one dimensional array of (possibly negative) floats representing reaction fluxes. Has same size as number of columns of self.structure.S.

        """  # Noqa: E501
        conc = get_conc(
            conc_balanced,
            self.parameters["log_conc_unbalanced"],
            self.structure,
        )
        flux_list = []
        for reaction_ix, (reaction_id, rate_equation) in enumerate(
            zip(self.structure.reactions, self.structure.rate_equations)
        ):
            ipt = rate_equation.get_input(
                parameters=self.parameters,
                reaction_id=reaction_id,
                reaction_stoichiometry=self.structure.S[:, reaction_ix],
                species_to_dgf_ix=self.structure.species_to_dgf_ix,
            )
            flux_list.append(rate_equation(conc, ipt))
        return jnp.array(flux_list)


class KineticModelSbml(KineticModel):
    sym_module: Any

    def flux(
        self,
        conc_balanced: Float[Array, " n_balanced"],
    ) -> Float[Array, " n"]:
        flux = jnp.array(
            self.sym_module(
                **self.parameters,
                **dict(zip(self.structure.balanced_species, conc_balanced)),
            )
        )
        return flux
